#!/usr/bin/perl

# barseq.pl main
# Last modification, 2012-05-01
#
# Not for general distribution.  The code in this library is under development.
# 
# Author : Lawrence Heisler
# Unversity of Toronto, HipHop Chemogenomics Laboratory
# http://chemogenomics.med.utoronto.ca
#
# Copyright (C) 2011 Lawrence Heisler <l.heisler@utoronto.ca>

use strict;
use warnings;
use barseq;
use Data::Dumper;

#### this simple implementation allows only three arguments, and uses many defaults
#### arg1 = file of reads, which must have either a fastq,or fq extenstion, or qseq.txt
#### arg2 = file of barcodes, which must have a list of barcodes
#### arg3 = dataset identification, the prefix for all data

my $argcount=scalar @ARGV;
my ($action,@args)=@ARGV;



if(! $action){
	print STDERR "first argument must be one of [count|cluster|map|tabulate\n";
	exit;
}

if($action eq "count"){
	my $readtype="";
	my $readfile=$args[0];
	if(! $readfile){
		print STDERR "readfile not indicated\n";
		exit;
	}
	elsif(! -e $readfile){
		print STDERR "readfile $readfile not found\n";
		exit;
	}
	elsif($readfile=~/qseq/i){
		$readtype="qseq";
	}	
	elsif($readfile=~/fastq|fq/){
		$readtype="fastq";
	}
	else{
		print STDERR "reads file must indicate type (qseq=qseq, fastq/fq=fastq)\n";
		exit;
	}
	
	print STDERR "counting reads in $readfile\n";
	my %tally=countReads($readfile,$readtype,10,0,0,"1:1",18);
	check_error(\%tally) || die "ERROR:unable to process\n";
	print STDERR "$tally{stats}{total} reads, $tally{stats}{qfiltered} filtered for quality, $tally{stats}{distinct} distinct sequences counted\n\n";	
	
	print_tally($tally{reads},$tally{stats});
	exit;
	
		
}
elsif($action eq "cluster"){
	### this requires a read file as the second argument
	
	my $countfile=$args[0];
	if(! $countfile){
		print STDERR "countfile not indicated\n";
		exit;
	}
	elsif(! -e $countfile){
		print STDERR "countfile $countfile not found\n";
		exit;
	}
	else{
		
		my %readcounts=load_reads($countfile);
		my %clusters=cluster_reads($readcounts{seq},0);
		check_error(\%clusters) || die "ERROR: unable to process\n";
		
		
		my $clusteredreads=$clusters{stats}{clusters}+$clusters{stats}{mm1clustered}+$clusters{stats}{mm2clustered};
		my $pcclustered=$clusters{stats}{readcount} ? $clusteredreads/$clusters{stats}{readcount}*100 : 0;
		print STDERR "$clusters{stats}{readcount} reads, $clusteredreads ($pcclustered %) assigned to $clusters{stats}{clusters} clusters (mm1=$clusters{stats}{mm1clustered}, mm2=$clusters{stats}{mm2clustered}) + $clusters{stats}{single} single read clusteres\n";
		
		print_clusters($clusters{cluster},$clusters{stats});
		
		
	}
	
	
}
elsif($action eq "map"){
	### requires two input files, 1. cluster file 2. barcodes
	my ($clusterfile,$barcodefile)=@args[0..1];
	if(! $clusterfile){
		print STDERR "file with clustered reads not indicated\n";
		exit;
	}
	elsif(! -e $clusterfile){
		print STDERR "file with clustered reads $clusterfile not found\n";
		exit;
	}
	if(! $barcodefile){
		print STDERR "file with expected barcodes not indicated\n";
		exit;
	}
	elsif(! -e $barcodefile){
		print STDERR "file with expected barcodes $barcodefile not found\n";
		exit;
	}
	
	### read in the barcodes
	my %bc=load_barcodes($barcodefile,18);
	check_error(\%bc) || die "ERROR: unable to process\n";
	print STDERR "$bc{count} barcodes loaded\n";

	my %clusters=load_clusters($clusterfile);
	my $clustercount=scalar keys %clusters;
	print STDERR "$clustercount clusters loaded\n";
	print STDERR "mapping barcodes to clusters\n";
	
	my %mapped=map_clusters(\%clusters,\%bc,1);   ### 1 flag indicates to do alignment
	check_error(\%mapped) || die "ERROR: unable to process\n";
	print STDERR "$mapped{stats}{matched} clusters matched, $mapped{stats}{aligned} aligned, $mapped{stats}{nomatch} not identified\n";
	print_map($mapped{cluster},$mapped{stats});
	exit;
}
elsif($action eq "tabulate"){
	my ($mapfile,$barcodefile)=@args[0..1];
	if(! $mapfile){
		print STDERR "file with mapped clusters not indicated\n";
		exit;
	}
	elsif(! -e $mapfile){
		print STDERR "file with mapped clusters $mapfile not found\n";
		exit;
	}
	if(! $barcodefile){
		print STDERR "file with expected barcodes not indicated\n";
		exit;
	}
	elsif(! -e $barcodefile){
		print STDERR "file with expected barcodes $barcodefile not found\n";
		exit;
	}
	
	
	### read in the barcodes
	my %bc=load_barcodes($barcodefile,18);
	check_error(\%bc) || die "ERROR: unable to process\n";
	print STDERR "$bc{count} barcodes loaded\n";
	
	my %mapped=load_map($mapfile);
	my $clustercount=scalar keys %mapped;
	print STDERR "$clustercount mapped clusters loaded\n";
	print STDERR "generating barcode list with counts and identities\n";
	
	my $align_minscore=75;
	print_mapped(\%mapped,\%bc,$align_minscore);
	
	
}
else{
	print STDERR "first argument must be one of [count|cluster|map|tabulate]";
	exit;
}


sub check_error{
	my ($object)=@_;
	if($$object{error}){
		print STDERR $$object{errorstring}."\n";
		return 0;
	}	
	return 1;
}

