# Library of functions for barseq analysis
# Last modification, 2012-05-01
#
# Not for general distribution.  The code in this library is under development.
# 
# Author : Lawrence Heisler
# Unversity of Toronto, HipHop Chemogenomics Laboratory
# http://chemogenomics.med.utoronto.ca
#
# Copyright (C) 2011 Lawrence Heisler <l.heisler@utoronto.ca>


package barseq;

use strict;
require Exporter;


use Data::Dumper;

our @ISA = qw(Exporter);

our @EXPORT = qw/countReads print_tally cluster_reads print_clusters load_clusters load_barcodes map_clusters 
				print_map print_mapped load_reads load_map/;






###################################################################################
### countReads : public function
### Parses through a qseq file and counts unique sequences
### sequences are trimmed if requested 
### based are marked as N if basecall quality does not exceed indicated cutoff 
### arg     : qseq file, quality cutoff, qkeep (keep or drop if filtered), read limit, trimming string
### calls   : get_read,qfilter,trim_read
### returns : tally
sub countReads{
    my ($file,$type,$qcutoff,$qkeep,$readlimit,$trimstring,$trunc)=@_;

    my ($trim5,$trim3)=split /:/,$trimstring;
    my %tally=(total=>0,pasfilter=>0,qfiltered=>0,distinct=>0);
    (open my $SEQ,"<",$file) || die "could not open seq file $file";
    while(!eof($SEQ))
    {
		last if($readlimit && $tally{total}>$readlimit);

		my %rec=get_read($SEQ,$type);
		
		my $errcount=$rec{error} ? scalar keys %{$rec{error}} : 0;
		if($errcount>0){
			$tally{errcount}=$errcount;
			$tally{errlist}=join(";",keys %{$rec{error}});
			return %tally;
		}

   
        $tally{stats}{total}++;
        $tally{stats}{passfilter}++ if($rec{pass});
        next unless($rec{pass});
        
        my $read=$rec{read};$read=~s/\./N/;   ## replace any . characters with N
        my $qual=$rec{qual};
        
        ### trim and truncate
        $read=trim_read($read,$trim5,$trim3,$trunc);
        $qual=trim_read($qual,$trim5,$trim3,$trunc);
        
        
        ### filter the read, based on quality
        my ($fread,$fcount)=$qcutoff? qfilter($read,$qual,$qcutoff,"I") : ($read,0);  ### filter read, replace lowQuality with N
        $tally{stats}{qfiltered}++ if($fcount>0);
        
        next if($fcount && !$qkeep);
        
        $tally{reads}{$fread}{count}++;    ### add to the hash
        
        
        

    }
    $tally{stats}{distinct}=scalar keys %{$tally{reads}};
    
    
    
    return(%tally);

}

###################################################################################
### print_tally : public function
### prints out the reads and counts.  Header is stats if supplied
### arg     : $reads - ref to hash of reads+counts; $stats - ref to hash of stats/values
### calls   : 
### returns : 
sub print_tally{
	my ($reads,$stats)=@_;
	
	my @statkeys=qw/total passfilter qfiltered distinct/;
	my $header=join("\n",map{"# ".$_.":".$$stats{$_}} @statkeys);
	print "$header\n";
	print "#sequence\tcount\n";
	
	for my $seq(sort{$$reads{$b}{count}<=>$$reads{$a}{count}} keys %$reads){
		my $count=$$reads{$seq}{count} || 0;
		print "$seq\t$count\n";
	}	


}

###################################################################################
### load_reads : public function
### loads reads and counts from a file.  first two columns must be 1. read sequence 2. count
### arg     : $file - the name of the file
### calls   : 
### returns : %counts - a hash of the counts
sub load_reads{
	my ($file)=@_;
	(open my $FH,"<",$file) || die "could not open file";
	my %counts;my $line;
	while(<$FH>){
		chomp;
		next if($_=~/^#/);
		$line++;
		my ($seq,$count,@info)=split /\t/;
		$counts{seq}{$seq}{count}=$count;
		
		### make sure seq is ACGTN only, make sure count is digits only,otherwise generate error
		if($seq=~/[^ACGTN]/i){
		    print STDERR "seq=$seq\n";
			$counts{errcount}++;
			$counts{errorlist}{"sequence at line $line has illegal characters, expecting only ACGTNacgtn"}++;
			return %counts;
		}	
		
		if($count=~/\D/){
			$counts{errcount}++;
			$counts{errorlist}{"count at line $line is non-numeric"}++;
			return %counts;
		}	
		
		
		
		
	}
	return %counts;	
}

###################################################################################
### get_read() : INTERNAL FUNCTION
### extracts a record from the sequence file, depending Parses through a qseq file and counts unique sequences
### arg     : $FH filehandle, for open sequence file; $type = "qseq" or "fastq"
### calls   : none
### returns : %rec hash, with keys (read,qual,pass,error)
sub get_read{
	my($FH,$type)=@_;
	my %rec;my $reccount;
	if($type eq "qseq"){
		$reccount++;
		my $qseqline=<$FH>;chomp($qseqline);
		
		my @f=split /\t/,$qseqline;
		@f==11 || $rec{error}{"incorrect number of fields in record $reccount of qseq file at record $reccount"}++;
		
		my ($read,$qual,$pass)=@f[8..10];

		$rec{error}{"read and qual strings of unequal length in qseq file at record $reccount"}++
				if(!$read || !$qual || length($read)!=length($qual));

        $rec{read}=$read;
        $rec{qual}=$qual;
        $rec{pass}=$pass;
        $rec{type}="qseq";
     }
     elsif($type eq "fastq"){
     	my $def1=<$FH>;
     	
     	if($def1!~/^@/){
     		$rec{error}{"invalid fastq file, defline does not begin with expected symbol \@"}
     	}
     	else{
     		 
    		my $read=<$FH>;chomp($read);
    		my $def2=<$FH>;
    		my $qual=<$FH>;chomp($qual);
    		
    		$rec{error}{"read and qual strings of unequal length is fastq file at record $reccount"}++
					if(!$read || !$qual || length($read)!=length($qual));
			
			$rec{read}=$read;
			$rec{qual}=$qual;
			$rec{pass}=1;  ### fastq files only include reads passing filter
			$rec{type}="fastq";
		}	
     
     
     }
     else{
     	$rec{error}{"unknown file type request"}++;
     }	
      
     return %rec;     
}

#####################################################
### qfilter() : INTERNAL FUNCTION to filter read based on quality scores
### arg     : read, quality string, cutoff, method
### calls   : 
### returns : read, filtercount
sub qfilter{
  my($read,$qual,$cutoff,$method)=@_;
  ## methods
  ## S - Sanger        Phred+33,  raw reads typically (0, 40)
  ## X - Solexa        Solexa+64, raw reads typically (-5, 40)
  ## I - Illumina 1.3+ Phred+64,  raw reads typically (0, 40)
  ## J - Illumina 1.5+ Phred+64,  raw reads typically (3, 40)

  my %params=( 
                S=>{lo=>0,hi=>40,adj=>33},
                X=>{lo=>(-5),hi=>40,adj=>64},
                I=>{lo=>0,hi=>40,adj=>64},
                J=>{lo=>3,hi=>40,adj=>64}
              );
  
  my @filter;
  for my $score($params{$method}{lo}..$params{$method}{hi})
  {
       my $val=$score+$params{$method}{adj};
       my $code=chr($val);
       #print "$score $val $code\n";
       push(@filter,$code) if($score<$cutoff);
  }
  
  my $filterstring=join("",@filter);  
  $filterstring=~s/([^A-Za-z0-9])/\\$1/g;
  #print "$filterstring\n";  

  my $filtercount=0;
  while($qual=~/[$filterstring]/g)
  {
     substr($read,$-[0],1,"N");
     $filtercount++;
  }
  return ($read,$filtercount);
}

########################################################################
### trim_read : INTERNAL FUNCTION TO TRIM READ
### arg     : read, trim5, trim3, truncate
### calls   : 
### returns : read
sub trim_read{
  my ($read,$trim5,$trim3,$trunc)=@_;
  $read=substr($read,$trim5);
  $read=substr($read,0,length($read)-$trim3);
  $read=substr($read,0,$trunc) if($trunc);
  return $read;
}

###################################################################################
### clusterReads : public function
### arg     : reference to reads hash, keys=read,value=count; $exclude : exclude reads seen less than number of times
### calls   : degenerate1
### returns : %cluster hash
sub cluster_reads{
	my($reads,$exclude)=@_;
	
	my @allseqs=sort {$$reads{$b}{count}<=>$$reads{$a}{count}} keys %$reads;
	my $allseqcount=scalar @allseqs;
	
	my @seqs=grep{$$reads{$_}{count}>$exclude} @allseqs;
	my $seqcount=scalar @seqs;
	
	print STDERR "clustering $seqcount sequences from $allseqcount, excluding sequence with count $exclude or less\n";
	
	my %hash;
	$hash{stats}={readcount=>0,clusters=>0,mm1clustered=>0,mm2clustered=>0,single=>0};
	
	#### SCORING : this is based on the number of times a mismatch is seen in the list of reads
	print STDERR "scoring sequences\n";
	for my $seq(@seqs){
		my $count=$$reads{$seq}{count};
		my $score=0;		
		my @mm=degenerate1($seq);
		map{$score++ if(defined($$reads{$_}) )} @mm;
		
		$hash{seq}{$seq}={count=>$count,score=>$score};
		@{$hash{seq}{$seq}{mm}}=@mm; 
		$hash{stats}{readcount}++;

 	}
	
	print STDERR "assigning core sequences\n";
	for my $seq(@seqs){
   		next if $hash{clustered}{$seq};   ### the read has been assigned to a cluster, it can't be a center
   		
   		my $score=$hash{seq}{$seq}{score} || 0;
   		my $count=$hash{seq}{$seq}{count} || 0;
   		
   		next unless($score && $count);   ### if the barcode has NO mismatches, then maybe it should be made into a center?  or should it be held aside to look for multiple mm
   		
   		my @mm=@{$hash{seq}{$seq}{mm}};
   		
   		### get the top mm....this code can be shortened
   		my $topmm;
   		my($topmmscore,$topmmcount)=(0,0);
   		for my $mm(@mm){
   			next unless($hash{seq}{$mm});
   			my $mmscore=$hash{seq}{$mm}{score};
   			my $mmcount=$hash{seq}{$mm}{count};
   			if($mmscore>$topmmscore){
   				if($mmcount>$topmmcount){
   					$topmm=$mm;$topmmscore=$mmscore;$topmmcount=$mmcount;
   				}	
   			}
   		}
   		
   		
   		
   		### linking barcodes to a center to form a cluster, none of the children have a higher parent
    	if($score>=$topmmscore){
         	## if a tie, only proceed if the higher counts
         	next if(($score == $topmmscore) && ($count<$topmmcount));  

         	### designate the read as a center
        	$hash{cluster}{$seq}{corecount}=$count;
        	$hash{cluster}{$seq}{score}=$score;
        	$hash{cluster}{$seq}{topmmscore}=$topmmscore;
        	
        	### indicate the read has been clustered
        	$hash{clustered}{$seq}={core=>$seq,count=>$count,mm=>"mm0"};
        	$hash{stats}{clusters}++;
         	
         	### indicate all the mms for the read have been clustered and add to the cluster
         	for my $mmseq(@mm){
        	    ### account for all mismatches, even if not in the list
        	    my $mmcount=$hash{seq}{$mmseq} ? $hash{seq}{$mmseq}{count} : 0;
        	    $hash{cluster}{$seq}{mm1}{$mmseq}=$mmcount;
        	    $hash{cluster}{$seq}{mm1count}+=$mmcount;
        	    
        	    $hash{clustered}{$mmseq}={center=>$seq,count=>$mmcount,mm=>"mm1"};  ## if mmcount=0, indicates NOT seen
        	    
        	    ### ONLY add to the mm1 count if a sequence in the dataset
        	    $hash{stats}{mm1clustered}++ if($hash{seq}{$mmseq});
        	    
        	    
        	}
        }

 	}
 	

 	print STDERR "expanding clusters\n";
 	### CLUSTER expansion
 	### determine if unclustered sequences can attach to a cluster by 2 mismatches
 	for my $seq(@seqs){
 		next if($hash{clustered}{$seq});
 		
 		my $count=$hash{seq}{$seq}{count} || 0;
		my $score=$hash{seq}{$seq}{score} || 0;
 		my @mm=@{$hash{seq}{$seq}{mm}};
 		my %centers;
		for my $mmseq(@mm){
			### has the mm been clustered
			my $center=$hash{clustered}{$mmseq} && ($hash{clustered}{$mmseq}{mm} eq "mm1") ? $hash{clustered}{$mmseq}{center} : 0;
			$centers{$center}++ if($center);

		}
		my @centers=keys %centers;
		my $centercount=scalar @centers;
		
		if($centercount){   ### the read will join a cluster
			$hash{stats}{mm2clustered}++;
			my $splitcount=int($count/$centercount);
			for my $center(keys %centers){
			
			    ######if(length($center>25)){print "bigcenter $center";<STDIN>;}
        	    $hash{cluster}{$center}{mm2}{$seq}=$splitcount;
        	    $hash{cluster}{$center}{mm2count}+=$splitcount;
        	}
			$hash{clustered}{$seq}={center=>join(";",@centers),count=>$count,mm=>"mm2"};
		}
		else{    ### the read will form its own cluster, unconnected to any other read
			#print "UNCLUSTERED read $seq";<STDIN>;
			$hash{stats}{single}++;
			$hash{cluster}{$seq}={corecount=>$count,score=>0,topmmscore=>0,mm1count=>0,mm2count=>0};
		}
	}
	


	

	
	return(%hash);
}		

###################################################################################
### print_clusters : public function
### arg     : reference to clusters hash, keys=cores; file to print; information to header
### calls   : 
### returns :
sub print_clusters{
	my ($clusters,$stats)=@_;
	
	
	my @statkeys=qw/readcount clusters mm1clustered mm2clustered single/;
	my $header=join("\n",map{"# ".$_.":".$$stats{$_}} @statkeys);
	print "$header\n";
	print "#center\ttotalcount\tscore\tmm0_total\tmm1_number\tmm1_total\tmm2_number\tmm2_total\tmm1_reads\tmm2_reads\n";
	
	my @cores=sort {$$clusters{$b}{corecount}<=>$$clusters{$a}{corecount}} keys %$clusters;
	my $corecount=scalar @cores;
	#print "saving $corecount clusters\n";
	for my $coreseq(@cores){
		my $core=$$clusters{$coreseq};

		
		my $score=$$core{score} || 0;
		my $topmmscore=$$core{topmmscore} || 0;
		
		my $mm0=$$core{corecount} || 0;
	 	my $mm1=$$core{mm1count} || 0;
	 	my $mm2=$$core{mm2count} || 0;
	 	my $total=$mm0+$mm1+$mm2;
	 	
	 	my @mm1s= map{$_.":".$$core{mm1}{$_}} sort{$$core{mm1}{$b}<=>$$core{mm1}{$a}} grep{$$core{mm1}{$_}} keys %{$$core{mm1}};
	 	my @mm2s= map{$_.":".$$core{mm2}{$_}} sort{$$core{mm2}{$b}<=>$$core{mm2}{$a}} grep{$$core{mm2}{$_}} keys %{$$core{mm2}};
	 	
	 	
	 	
	 	my $mm1n=scalar @mm1s;
	 	my $mm2n=scalar @mm2s;
		
		print  "$coreseq\t$total\t$score:$topmmscore\t$mm0\t$mm1n\t$mm1\t$mm2n\t$mm2\t".join(",",@mm1s)."\t".join(",",@mm2s)."\n"; 

	}

}

###################################################################################
### read_clusters : public function
### arg     : filename, from which clusters should be read.
### calls   : get_words
### returns : %bc a hash, keys=seq (nmer sequences)->ids, ids=ids->seq, words->seq->count
sub load_clusters{
	my ($file)=@_;
	my %clusterdata;
	
	(open my $FH,$file) || die "could not open file";
	while(<$FH>){
		chomp;
		next if($_=~/^#/);

		my($core,$total,$score,$mm0,$mm1n,$mm1,$mm2n,$mm2,$mm1seqs,$mm2seqs)=split/\t/;
		$clusterdata{$core}={score=>$score,count=>$total,mm0=>$mm0,mm1=>$mm1,mm2=>$mm2,
								mm1n=>$mm1n,mm2n=>$mm2n,mm1seqs=>$mm1seqs,mm2seqs=>$mm2seqs};
	
		
						
	}
	return (%clusterdata);
	

}

###################################################################################
### load_barcodes : public function
### arg     : filename: file is a list of barcodes + ids.  barcode sequences should be in the correct orientation
### arg		: the length of the barcode sequence , should match trimming of reads
### calls   : 
### returns :
sub load_barcodes{
	my ($file,$len)=@_;
	my %bc;
	my $BC;
	unless(open $BC,"<",$file){
		$bc{error}++;
		$bc{errorstring}="barcode file $file not found";
		return %bc;
	}
	my $line=0;
	while(<$BC>){
		$line++;
		chomp;
		my @f=split /\t/;
		if(scalar @f != 2){
			$bc{error}++;
			$bc{errorstring}="found " . scalar @f . " fields at line $line of barcode file $file";
			return %bc;
		} 
		
		my($id,$barcode)=split /\t/;
		### check for probe letters
		if($barcode=~/[^ACGTN]/i){
			$bc{error}++;
			$bc{errorstring}="barcode sequence at line $line has illegal characters, expecting only ACGTNacgtn";
			return %bc;
		}
		if(exists($bc{ids}{$id})){
			$bc{error}++;
			$bc{errorstring}="duplicate barcode id $id found at line $line";
			return %bc;
		}
		
		$bc{ids}{$id}=$barcode;
		$bc{fullseq}{$barcode}=$id;
		
		### check that barcode length is at least as long as $len
		$len=length($barcode) unless($len);
		for my $i(0..length($barcode)-$len){
			my $seq=substr($barcode,$i,$len);
			$bc{seq}{$seq}{$id}++;
		}
		
		my @words=get_words($barcode,4);
		for my $word(@words){
			$bc{words}{$word}{$barcode}++;
		}	
		
		
	}
	$bc{count}=scalar keys %{$bc{ids}};
	
	
	return %bc;
}

###################################################################################
### map_clusters : public function
### arg     : clusters hash,barcodes hash, align flag
### calls   : get_words,align
### returns : %map : hash of mappings
sub map_clusters{
	my ($clusters,$barcodes,$align)=@_;

	my %map;
	my @coreseqs=sort {$$clusters{$b}{count}<=>$$clusters{$a}{count}} keys %$clusters;
	
	$map{stats}={clusters=>0,matched=>0,"match-mm0"=>0,"match-mm1"=>0,"match-mm2"=>0,nomatch=>0,aligned=>0};
	
	for my $coreseq(@coreseqs){

   		$map{stats}{clusters}++;
   		
   		@{$map{cluster}{$coreseq}}{qw/score maxmmscore count mm0 mm1 mm2 mm1n mm2n/}= @{$$clusters{$coreseq}}{qw/score maxmmscore count mm0 mm1 mm2 mm1n mm2n/};
		
		my %mm=(mm0=>{$coreseq=>$$clusters{$coreseq}{mm0}});
		%{$mm{mm1}}=map{split /:/,$_} split(/,/,$$clusters{$coreseq}{mm1seqs});
		%{$mm{mm2}}=map{split /:/,$_} split(/,/,$$clusters{$coreseq}{mm2seqs});
		
		my %hits;
		for my $match(qw/mm0 mm1 mm2/){    #### EXPECTATION is only ONE HIT
			my ($seq,$id)= map{$_,join(";",keys %{$$barcodes{seq}{$_}}) } grep($$barcodes{seq}{$_}, keys %{$mm{$match}});
			
			$hits{$id}={match=>$match,seq=>$seq,count=>$mm{$match}{$seq}} if($id);
		}
		my $id=(keys %hits)[0];
		
		if($id){
			@{$map{cluster}{$coreseq}}{qw/match seq count/}=@{$hits{$id}}{qw/match seq count/};
			$map{cluster}{$coreseq}{id}=$id;
			$map{cluster}{$coreseq}{seq}=$$barcodes{ids}{$id} || "";
			
			$map{stats}{matched}++;
			my $matchtype="match-".$map{cluster}{$coreseq}{match};
			$map{stats}{$matchtype}++;


		}
		elsif($align){
			#print "aligning $coreseq";
			my @words=get_words($coreseq,4);
			
			my $minshare=6;
			my %share;
			map{$share{$_}++} map{keys %$_} @{$$barcodes{words}}{@words};
			
			my $bestshare=(sort{$b<=>$a} values %share)[0];
			if($bestshare>=$minshare)
			{
				my @candidates=grep {$share{$_}==$bestshare} keys %share;
				my %aligns;
			
				for my $candseq(@candidates){
					my %alignment=align($candseq,$coreseq,5,-1,-1);
					#print STDERR Dumper(%alignment);<STDIN>;
					my $alignscore=$alignment{score};
					%{$aligns{$alignscore}{$candseq}}=%alignment;
				}
				my $maxalignscore=(sort{$b<=>$a} keys %aligns)[0];
				my @alignedseqs=keys %{$aligns{$maxalignscore}};
				my $aligncount=scalar @alignedseqs;
				my $id=join(";",@{$$barcodes{fullseq}}{@alignedseqs});
	

				$id=$aligncount>1 ? "MULTIPLE:$id" : $id;

				$map{cluster}{$coreseq}{id}=$id;
				$map{cluster}{$coreseq}{match}="align";
				$map{cluster}{$coreseq}{seq}=join(";",@alignedseqs);
				
				$map{stats}{aligned}++;
				$map{cluster}{$coreseq}{aligns}{score}=$maxalignscore;
				
				my $revised=join(";",map{$aligns{$maxalignscore}{$_}{j}} @alignedseqs);
				my $mid=join(";",map{ $aligns{$maxalignscore}{$_}{mismatch}.":".
									  $aligns{$maxalignscore}{$_}{in}      .":".
									  $aligns{$maxalignscore}{$_}{del} 
									} @alignedseqs);
				$map{cluster}{$coreseq}{aligns}{revision}=$revised;
				$map{cluster}{$coreseq}{aligns}{mid}=$mid;

			}
			else{
				$map{cluster}{$coreseq}{match}="none";
				$map{stats}{nomatch}++;
			}		
			

		}
		else{
			$map{cluster}{$coreseq}{match}="none";
			$map{stats}{nomatch}++;
		}
	}
	return %map;

}

###################################################################################
### align : internal function, implementation of dynamic programming algorithm
### arg     : seqi,seqj : sequences to align
### arg 	: scoring : match, mm, gap
### calls   : 
### returns : %align : hash of alignment
sub align{
	
	
	my($seqi,$seqj,@scorescheme)=@_;
	###print "aligning $seqi to $seqj\n";
	my %scoring;
	
	
	@scoring{qw/match mismatch gap/}=@scorescheme;

	my @ibases=split //,$seqi;
	my @jbases=split //,$seqj;
	
	my @matrix;
	$matrix[0][0]={score=>0,source=>"none"};
	for my $i(1..@ibases){$matrix[$i][0]={score=>$scoring{gap}*$i,source=>"i"}}
	for my $j(1..@jbases){$matrix[0][$j]={score=>$scoring{gap}*$j,source=>"j"}}
	
	#draw_matrix($seqi,$seqj,\@matrix);<STDIN>;
	
	for my $i(1..@ibases){
		my $ibase=$ibases[$i-1];
		for my $j(1..@jbases){
			my $jbase=$jbases[$j-1];
			
			my %scores;
			my $mod=$ibase eq $jbase ? $scoring{match} : $scoring{mismatch};
	
			$scores{ij}=$matrix[$i-1][$j-1]{score}+$mod;
			$scores{i}=$matrix[$i-1][$j]{score} + $scoring{gap};
			$scores{j}=$matrix[$i][$j-1]{score} + $scoring{gap};
			
			#ßprint Dumper(%scores);<STDIN>;
			
			my @orderedscores=sort {$b<=>$a} values %scores;
			my $bestscore=$orderedscores[0];
			$matrix[$i][$j]{score}=$bestscore;
			my $source;
			if($scores{ij}==$bestscore){$source="ij"}elsif($scores{i}==$bestscore){$source="i"}else{$source="j"}
			$matrix[$i][$j]{source}=$source;
			
			
		}
	}	
	#draw_matrix($seqi,$seqj,\@matrix);<STDIN>;	
	
	
	my %alignment=(in=>0,del=>0,mismatch=>0);

	# start at last cell of matrix
	my $i = length($seqi);
	my $j = length($seqj);

	my $source=$matrix[$i][$j]{source};
	$alignment{score}=$matrix[$i][$j]{score};
	
	my %aligns;

	while ($source ne "none") {
	  	my $ibase=$ibases[$i-1];
	  	my $jbase=$jbases[$j-1];
		if ($source eq "ij") {
        	push(@{$aligns{i}},$ibase);
        	push(@{$aligns{j}},$jbase);
        	$i--;
        	$j--;
        	$alignment{mismatch}++ if($ibase ne $jbase);
    	}elsif ($source eq "i") {
        	push(@{$aligns{i}},$ibase);
        	push(@{$aligns{j}},"-");
        	$i--;
        	$alignment{in}++;
    	}elsif ($source eq "j") {
        	push(@{$aligns{i}},"-");
        	push(@{$aligns{j}},$jbase);
        	$j--;
        	$alignment{del}++;
		} 
    	$source=$matrix[$i][$j]{source};
    
	}
	$alignment{i}=join("",reverse(@{$aligns{i}}));
	$alignment{j}=join("",reverse(@{$aligns{j}}));
	
	return %alignment;
}

###################################################################################
### print_map 	: public function
### arg     : map hash, file to print to 
### calls   : 
### returns : %align : hash of alignment
sub print_map{
	my ($map,$stats)=@_;
	
	my @statkeys=qw/clusters matched match-mm0 match-mm1 match-mm2 aligned nomatch/;
	my $header=join("\n",map{"# ".$_.":".$$stats{$_}} @statkeys);
	print "$header\n";
	my @fields=qw/count score mm0 mm1n mm1 mm2n mm2 match id seq/;
	print "#core\t".join("\t",@fields) . "\talignments\n";
	my @coreseqs=sort{$$map{$b}{count}<=>$$map{$a}{count}} keys %{$map};
	for my $coreseq(@coreseqs){
		my $counts=$$map{$coreseq}{counts};
		my @values=@{$$map{$coreseq}}{@fields};
		my $aligninfo=$$map{$coreseq}{match} eq "align" ? $$map{$coreseq}{aligns}{score}.":".$$map{$coreseq}{aligns}{revision}.":".$$map{$coreseq}{aligns}{mid} : "";
		
		
		
		print "$coreseq\t".join("\t",@values)."\t$aligninfo \n";
	}

}

sub load_map{
	my ($file)=@_;
	my %mapdata;

	(open my $FH,$file) || die "could not open file";
	while(<$FH>){
		chomp;
		next if($_=~/^#/);

		my($core,$count,$score,$mm0,$mm1n,$mm1,$mm2n,$mm2,$match,$id,$seq,$align)=split/\t/;
		$mapdata{$core}={score=>$score,count=>$count,mm0=>$mm0,mm1=>$mm1,mm2=>$mm2,
								mm1n=>$mm1n,mm2n=>$mm2n,match=>$match,id=>$id,seq=>$seq,align=>$align};
	}
	return (%mapdata);
}

###################################################################################
### print_mapped 	: public function
### arg     : map hash, barcode hash, file to print to 
### calls   : 
### returns : %align : hash of alignment
sub print_mapped{
	my($map,$bc,$minscore)=@_;
		
	my @coreseqs=sort{$$map{$b}{count}<=>$$map{$a}{count}} keys %{$map};
	
	
	my %bccounts;my %stats;
	### tabulate counts for each barcode
	print STDERR "tabulating barcode counts from mapping\n";
	for my $coreseq(@coreseqs){
		
		my $id=$$map{$coreseq}{id} || 0;
		
		if($id){
			my $match=$$map{$coreseq}{match};
			
			if($match eq "align"){
			
			    my($alignscore,$alignseq,$alignmid)=split /:/,$$map{$coreseq}{align};
				if($alignscore>=$minscore){
					if($id=~/MULTIPLE/i){
			    		$stats{ambiguous}++;
			    	}
			    	else{
			    		
			    		
			    		$bccounts{$id}{counts}{total}+=$$map{$coreseq}{count};
			    		$bccounts{$id}{counts}{$match}+=$$map{$coreseq}{count};
			    		
			    		
			    		my $revision=$$map{$coreseq}{align};
			    		$bccounts{$id}{counts}{revisions}{$revision}++;
			    	}
				}
				else{
					$stats{weakalign}++;
				}
			}
			else{
				$bccounts{$id}{counts}{total}+=$$map{$coreseq}{count};
				$bccounts{$id}{counts}{$match}+=$$map{$coreseq}{count};
			}	
		}
	}

	print STDERR "generating barcode list with counts\n";
	my @ids=sort keys %{$$bc{ids}};
	
	print "bcid\tbarcode\ttotal\tmm0\tmm1\tmm2\talign\n";
	for my $id(@ids){
		my $barcode=$$bc{ids}{$id};
		
		my($total,$mm0,$mm1,$mm2,$align)=map{$_?$_:0} @{$bccounts{$id}{counts}}{qw/total mm0 mm1 mm2 align/};
	    $total=0 unless($total);
	    my $revisions=$align>0 ? join(";",keys %{$bccounts{$id}{counts}{revisions}}) : "";
		print "$id\t$barcode\t$total\t$mm0\t$mm1\t$mm2\t$align\t$revisions\n";
	}		

}

###################################################################################
### degenerate1 	: internal function
### arg     : seq : sequence to degenerate 
### calls   : 
### returns : @degenerates : array of all single base degenerates
sub degenerate1{
	my ($seq)=@_;
	
	my %degenerates;
	
	my @bases=split //,$seq;
	for my $pos(0..$#bases)
	{
	   my $base=$bases[$pos];
	   #print "$pos $base";<STDIN>;
	   for my $alt(qw/A C G T N/)
	   {
	   		next if($alt eq $base);
	   		my $degen=$seq;
	   		substr($degen,$pos,1,$alt);
	   		$degenerates{$degen}++;
        }	   
	}
	my @degenerates=sort keys %degenerates;
	return @degenerates;
	

}

###################################################################################
### get_words 	: internal function
### arg    	: seq=sequence from which to extract words, wl length of word
### calls   : 
### returns : @words : array of all words in the sequences
sub get_words{
   my($seq,$wl)=@_;
   my %uniquewords;
   for my $start(0..length($seq)-$wl){
      my $word=substr($seq,$start,$wl);
      $uniquewords{$word}++;
   }
   my @words=keys %uniquewords;
   return @words;
}

###########################################
### rc 		: internal function, reverse/complements a sequences
### arg     : read, act(x,r,c,rc)
### calls   : 
### returns : adjusted read sequence
sub rc
{
  my($seq,$act)=@_;
  $seq=reverse($seq) if($act=~/r/i);
  $seq=~tr/acgtACGT/tgcaTGCA/ if($act=~/c/i);
  return($seq);
}










###################################################
1;
